#include <stdio.h>
#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "elf.h"
#include <capstone/capstone.h>

#define MIN(a, b) ((a) < (b) ? (a) : (b))

struct list
{
    struct list *next;
    struct list *prev;
};

struct elf
{
    uint8_t *p_buf;
    size_t n_buf;

    size_t phent_off;
    size_t phent_count;
    size_t phent_size;

    size_t shent_off;
    size_t shent_count;
    size_t shent_size;

    uint32_t entry;

    size_t sh_strtab;
};

struct seg
{
    uint8_t *ptr;
    uint32_t base;
    uint32_t limit;
    uint32_t size;
    struct list node;
};

enum reg
{
    REG_EAX = 0,
    REG_ECX = 1,
    REG_EDX = 2,
    REG_EBX = 3,
    REG_ESP = 4,
    REG_EBP = 5,
    REG_ESI = 6,
    REG_EDI = 7
};

struct rel
{
    uint32_t offset;
    const char *name;
    struct list node;
};

struct emu
{
    int run;

    struct list ram;
    struct list relocs;

    uint32_t eip;
    uint32_t next_eip;
    uint32_t gpr[8];

    csh disasm;
    cs_insn *insn;

    uint32_t sigill_handler;
    uint32_t sigseg_handler;

    int      seg_fail;
    uint32_t seg_address;

    uint32_t seg_alloc_base;
};

static int error(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
     fprintf(stdout, "ERROR: ");
    vfprintf(stdout, fmt, args);
     fprintf(stdout, "\n");
     fflush(stdout);
    va_end(args);

    return EXIT_FAILURE;
}

static int msg(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
     fprintf(stdout, "; ");
    vfprintf(stdout, fmt, args);
     fprintf(stdout, "\n");
     fflush(stdout);
    va_end(args);

    return EXIT_FAILURE;
}

static uint32_t align(uint32_t address, uint32_t boundary)
{
    uint32_t bits = boundary - 1;
    return (address + bits) & ~bits;
}

static void list_init(struct list *head)
{
    head->next = head;
    head->prev = head;
}

static void list_append(struct list *head, struct list *node)
{
    struct list *prev;

    prev = head->prev;

    node->next = head;
    node->prev = prev;
    prev->next = node;
    head->prev = node;
}

static void list_remove(struct list *node)
{
    struct list *next;
    struct list *prev;

    next = node->next;
    prev = node->prev;

    prev->next = next;
    next->prev = prev;
}

static inline void *list_enum(struct list *head, void **it, size_t offset)
{
    void *obj = NULL;
    struct list *node = *it;

    if (node == NULL)
        node = head->next;

    if (node == head)
    {
        *it = NULL;
        return NULL;
    }

    obj = (void *) ((size_t) node - offset);
    *it = node->next;

    return obj;
}

static int elf_parse(struct elf *elf, uint8_t *buf, size_t len)
{
    const Elf32_Ehdr *eh = NULL;

    if (len < sizeof(Elf32_Ehdr))
        return EXIT_FAILURE;

    eh = (Elf32_Ehdr *) buf;

    if (eh->e_ident[EI_MAG0] != ELFMAG0 ||
        eh->e_ident[EI_MAG1] != ELFMAG1 ||
        eh->e_ident[EI_MAG2] != ELFMAG2 ||
        eh->e_ident[EI_MAG3] != ELFMAG3 ||
        eh->e_ident[EI_CLASS] != ELFCLASS32 ||
        eh->e_ident[EI_DATA] != ELFDATA2LSB ||
        eh->e_ident[EI_VERSION] != EV_CURRENT ||
        eh->e_ident[EI_OSABI] != ELFOSABI_SYSV ||
        eh->e_type != ET_EXEC ||
        eh->e_machine != EM_386 ||
        eh->e_version != EV_CURRENT)
    {
        return EXIT_FAILURE;
    }

    if (eh->e_phoff + eh->e_phnum * eh->e_phentsize > len ||
        eh->e_shoff + eh->e_shnum * eh->e_shentsize > len)
    {
        return EXIT_FAILURE;
    }

    memset(elf, 0, sizeof(*elf));

    elf->p_buf = buf;
    elf->n_buf = len;

    elf->phent_off = eh->e_phoff;
    elf->phent_count = eh->e_phnum;
    elf->phent_size = eh->e_phentsize;

    elf->shent_off = eh->e_shoff;
    elf->shent_count = eh->e_shnum;
    elf->shent_size = eh->e_shentsize;

    elf->sh_strtab = eh->e_shstrndx;

    elf->entry = eh->e_entry;

    return EXIT_SUCCESS;
}

static const Elf32_Shdr *elf_get_section(struct elf *elf, size_t i)
{
    return (const Elf32_Shdr *) (elf->p_buf + elf->shent_off + i * elf->shent_size);
}

static const void *elf_get_segment_ptr(struct elf *elf, const Elf32_Phdr *ph)
{
    return (const void *) (elf->p_buf + ph->p_offset);
}

static const void *elf_get_section_ptr(struct elf *elf, const Elf32_Shdr *sh)
{
    return (const void *) (elf->p_buf + sh->sh_offset);
}

static int elf_enum_segments(struct elf *elf, int *it, const Elf32_Phdr **o_ph)
{
    int i = *it;

    if (i == (int) elf->phent_count)
    {
        *it = 0;
        return 0;
    }

    *o_ph = (Elf32_Phdr *) (elf->p_buf + elf->phent_off + i * elf->phent_size);
    *it += 1;

    return 1;
}

static int elf_enum_sections(struct elf *elf, int *it, const Elf32_Shdr **o_sh)
{
    int i = *it;

    if (*it == (int) elf->shent_count)
    {
        *it = 0;
        return 0;
    }

    *o_sh = elf_get_section(elf, (size_t) i);
    *it += 1;

    return 1;
}

static int elf_enum_reloc(struct elf *elf,
                          const Elf32_Shdr *reltab,
                          int *it, Elf32_Rela *o_reloc)
{
    int i = *it;
    const void *table;

    assert(reltab->sh_type == SHT_REL ||
           reltab->sh_type == SHT_RELA);

    if (*it == (int) (reltab->sh_size / reltab->sh_entsize))
    {
        *it = 0;
        return 0;
    }

    table = elf_get_section_ptr(elf, reltab);

    if (reltab->sh_type == SHT_REL)
    {
        const Elf32_Rel *rel = table;;
        o_reloc->r_offset = rel[i].r_offset;
        o_reloc->r_info = rel[i].r_info;
        o_reloc->r_addend = 0;
    }
    else if (reltab->sh_type == SHT_RELA)
    {
        const Elf32_Rela *rela = table;
        o_reloc->r_offset = rela[i].r_offset;
        o_reloc->r_info = rela[i].r_info;
        o_reloc->r_addend = rela[i].r_addend;
    }
    else
        return 0;

    *it += 1;

    return 1;
}

static int elf_enum_symbols(struct elf *elf,
                            const Elf32_Shdr *symtab,
                            int *it, const Elf32_Sym **o_sym)
{
    int i = *it;
    const void *table;

    assert(symtab->sh_type == SHT_SYMTAB ||
           symtab->sh_type == SHT_DYNSYM);

    if (*it == (int) (symtab->sh_size / symtab->sh_entsize))
    {
        *it = 0;
        return 0;
    }

    table = elf_get_section_ptr(elf, symtab);
    *o_sym = (Elf32_Sym *) table + i;

    *it += 1;

    return 1;
}

static const char *elf_get_string(struct elf *elf, size_t strtab, size_t offset)
{
    const Elf32_Shdr *sh = elf_get_section(elf, strtab);
    return (const char *) elf_get_section_ptr(elf, sh) + offset;
}

static const Elf32_Sym *elf_get_symbol(struct elf *elf, const Elf32_Shdr *symtab, size_t i)
{
    return (const Elf32_Sym *) elf_get_section_ptr(elf, symtab) + i;
}

static int seg_init(struct seg *s, uint32_t size, uint32_t base)
{
    if ((s->ptr = malloc(size)) == NULL)
        return EXIT_FAILURE;

    s->base = base;
    s->limit = base + size;
    s->size = size;

    return EXIT_SUCCESS;
}

static void seg_free(struct seg *s)
{
    free(s->ptr);

    s->ptr = NULL;
    s->base = 0;
    s->size = 0;
}

static int seg_get_offset(struct seg *s, uint32_t va,
                          size_t count, size_t *p_off)
{
    if (va < s->base || va >= s->limit + count)
        return EXIT_FAILURE;

    *p_off = va - s->base;

    return EXIT_SUCCESS;
}

static int seg_read(struct seg *s, uint32_t va, size_t count, uint8_t **p_buf)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, count, &offset)) != EXIT_SUCCESS)
        return ret;

    *p_buf = s->ptr + offset;

    return EXIT_SUCCESS;
}

static int seg_read_byte(struct seg *s, uint32_t va, uint32_t *p_byte)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, 1, &offset)) != EXIT_SUCCESS)
        return ret;

    *p_byte = s->ptr[offset];

    return EXIT_SUCCESS;
}

static int seg_read_half(struct seg *s, uint32_t va, uint32_t *p_half)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, 2, &offset)) != EXIT_SUCCESS)
        return ret;

    *p_half = ((uint16_t) s->ptr[offset + 0] << 0) |
              ((uint16_t) s->ptr[offset + 1] << 8);

    return EXIT_SUCCESS;
}

static int seg_read_word(struct seg *s, uint32_t va, uint32_t *p_word)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, 4, &offset)) != EXIT_SUCCESS)
        return ret;

    *p_word = ((uint32_t) s->ptr[offset + 0] << 0) |
              ((uint32_t) s->ptr[offset + 1] << 8) |
              ((uint32_t) s->ptr[offset + 2] << 16) |
              ((uint32_t) s->ptr[offset + 3] << 24);

    return EXIT_SUCCESS;
}

static int seg_write_byte(struct seg *s, uint32_t va, uint32_t byte)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, 1, &offset)) != EXIT_SUCCESS)
        return ret;

    s->ptr[offset] = byte;

    return EXIT_SUCCESS;
}

static int seg_write_half(struct seg *s, uint32_t va, uint32_t half)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, 2, &offset)) != EXIT_SUCCESS)
        return ret;

    s->ptr[offset + 0] = (half >> 0) & 0xff;
    s->ptr[offset + 1] = (half >> 8) & 0xff;

    return EXIT_SUCCESS;
}

static int seg_write_word(struct seg *s, uint32_t va, uint32_t word)
{
    size_t offset = 0;
    int ret = EXIT_FAILURE;

    if ((ret = seg_get_offset(s, va, 4, &offset)) != EXIT_SUCCESS)
        return ret;

    s->ptr[offset + 0] = (word >> 0) & 0xff;
    s->ptr[offset + 1] = (word >> 8) & 0xff;
    s->ptr[offset + 2] = (word >> 16) & 0xff;
    s->ptr[offset + 3] = (word >> 24) & 0xff;

    return EXIT_SUCCESS;
}

static struct seg *emu_add_segment(struct emu *e, size_t size, uint32_t base)
{
    struct seg *s;
    int ret = EXIT_FAILURE;

    if ((s = malloc(sizeof(*s))) == NULL)
        return NULL;

    if ((ret = seg_init(s, size, base)) != EXIT_SUCCESS)
    {
        free(s);
        return NULL;
    }

    list_append(&e->ram, &s->node);

    return s;
}

static struct seg *emu_find_segment(struct emu *e, uint32_t va)
{
    struct list *node = NULL;
    struct list *head = &e->ram;

    node = head->next;

    while (node != head)
    {
        struct seg *s = (struct seg *)
            ((size_t) node - offsetof(struct seg, node));

        if (va >= s->base && va < s->limit)
            return s;

        node = node->next;
    }

    return NULL;
}

static int emu_load_segment(struct emu *e, struct elf *elf,
                            const Elf32_Phdr *ph, size_t extra)
{
    struct seg *s = NULL;
    int ret = EXIT_SUCCESS;

    msg("allocating segment: base=%.8x, limit=%.8x, flags: %x",
        ph->p_vaddr, ph->p_vaddr + ph->p_memsz, ph->p_flags);

    if ((s = emu_add_segment(e, ph->p_memsz + extra, ph->p_vaddr)) == NULL)
        return ret;

    memcpy(s->ptr, elf_get_segment_ptr(elf, ph), ph->p_filesz);

    return EXIT_SUCCESS;
}

static int emu_init_relocs(struct emu *e, struct elf *elf)
{
    int i = 0;
    const Elf32_Shdr *sh = NULL;
    struct rel *rel = NULL;

    list_init(&e->relocs);

    while (elf_enum_sections(elf, &i, &sh))
    {
        if (sh->sh_type == SHT_REL || sh->sh_type == SHT_RELA)
        {
            int j = 0;
            Elf32_Rela re;

            while (elf_enum_reloc(elf, sh, &j, &re))
            {
                size_t i_sym = ELF32_R_SYM(re.r_info);
                const Elf32_Shdr *symtab = elf_get_section(elf, sh->sh_link);
                const Elf32_Sym *sym = elf_get_symbol(elf, symtab, i_sym);

                if ((rel = malloc(sizeof(*rel))) == NULL)
                    return EXIT_FAILURE;

                rel->offset = re.r_offset;
                rel->name = elf_get_string(elf, symtab->sh_link, sym->st_name);
                list_append(&e->relocs, &rel->node);
            }
        }
    }

    return EXIT_SUCCESS;
}

static void emu_free_seg(struct seg *s)
{
    seg_free(s);
    free(s);
}

static void emu_free_ram(struct emu *e)
{
    void *i = NULL;
    struct seg *s = NULL;

    while ((s = list_enum(&e->ram, &i, offsetof(struct seg, node))) != NULL)
        emu_free_seg(s);
}

static void emu_free_relocs(struct emu *e)
{
    void *i = NULL;
    struct rel *r = NULL;

    while ((r = list_enum(&e->relocs, &i, offsetof(struct rel, node))) != NULL)
    {
        free(r);
    }
}

static int emu_init(struct emu *e, struct elf *elf)
{
    int i = 0;
    int ret = EXIT_SUCCESS;
    const Elf32_Phdr *ph = NULL;
    const uint32_t stack_base = 0xf0000000;
    const uint32_t stack_size = 1 * 1024 * 1024;

    memset(e, 0, sizeof(*e));

    if (cs_open(CS_ARCH_X86, CS_MODE_32, &e->disasm) != CS_ERR_OK)
        return EXIT_FAILURE;

    if (cs_option(e->disasm, CS_OPT_DETAIL, CS_OPT_ON) != CS_ERR_OK)
        goto cleanup;

    if ((e->insn = cs_malloc(e->disasm)) == NULL)
        goto cleanup;

    list_init(&e->ram);

    if (emu_add_segment(e, stack_size, stack_base) == NULL)
        goto cleanup;

    while (elf_enum_segments(elf, &i, &ph))
    {
        if (ph->p_type == PT_LOAD)
            emu_load_segment(e, elf, ph, 32);
    }

    if ((ret = emu_init_relocs(e, elf)) != EXIT_SUCCESS)
    {
        emu_free_ram(e);
        goto cleanup;
    }

    e->gpr[REG_ESP] = stack_base + stack_size;
    e->eip = elf->entry;

    e->seg_alloc_base = 0x10000000;

    return EXIT_SUCCESS;

cleanup:
    if (e->insn != NULL)
        cs_free(e->insn, 1);

    cs_close(&e->disasm);
    return EXIT_FAILURE;
}

static void emu_free(struct emu *e)
{
    emu_free_ram(e);
    emu_free_relocs(e);

    cs_free(e->insn, 1);
    cs_close(&e->disasm);
}

static struct rel *emu_find_reloc(struct emu *e, uint32_t offset)
{
    void *i = NULL;
    struct rel *r = NULL;

    while ((r = list_enum(&e->relocs, &i, offsetof(struct rel, node))) != NULL)
    {
        if (r->offset == offset)
            return r;
    }

    return NULL;
}

static uint32_t emu_read_reg(struct emu *e, enum reg n, int part)
{
    uint32_t v = e->gpr[n];

    switch (part & 3)
    {
        case 0: return v & 0x000000ff;
        case 1: return v & 0x0000ff00;
        case 2: return v & 0x0000ffff;
        default: return v;
    }
}

static uint32_t emu_write_reg(struct emu *e, enum reg n, int part, uint32_t word)
{
    uint32_t v = e->gpr[n];

    switch (part & 3)
    {
        case 0: return e->gpr[n] = (v & 0xffffff00) | word;
        case 1: return e->gpr[n] = (v & 0xffff00ff) | word << 8;
        case 2: return e->gpr[n] = (v & 0xffff0000) | word;
        default: return e->gpr[n] = word;
    }
}

static int emu_get_mem_ptr(struct emu *e, uint32_t va, int size, uint8_t **o_ptr)
{
    struct seg *s = emu_find_segment(e, va);

    if (s == NULL)
        return EXIT_FAILURE;

    return seg_read(s, va, size, o_ptr);
}

static int emu_read_mem(struct emu *e, uint32_t va, int size, uint32_t *o_data)
{
    int ret = EXIT_FAILURE;
    struct seg *s = emu_find_segment(e, va);

    if (s != NULL)
    {
        switch (size)
        {
            case 1: ret = seg_read_byte(s, va, o_data); break;
            case 2: ret = seg_read_half(s, va, o_data); break;
            case 4: ret = seg_read_word(s, va, o_data); break;

            default:
                return EXIT_FAILURE;
        }
    }

    if (ret == EXIT_FAILURE)
    {
        e->seg_fail = 1;
        e->seg_address = va;
    }

    return ret;
}

static int emu_write_mem(struct emu *e, uint32_t va, int size, uint32_t data)
{
    int ret = EXIT_SUCCESS;
    struct seg *s = emu_find_segment(e, va);

    if (s != NULL)
    {
        switch (size)
        {
            case 1: ret = seg_write_byte(s, va, data); break;
            case 2: ret = seg_write_half(s, va, data); break;
            case 4: ret = seg_write_word(s, va, data); break;

            default:
                return EXIT_FAILURE;
        }
    }

    if (ret == EXIT_FAILURE)
    {
        e->seg_fail = 1;
        e->seg_address = va;
    }

    return ret;
}

static uint32_t emu_read_reg_cs(struct emu *e, x86_reg n)
{
    switch (n)
    {
        case X86_REG_EAX: return emu_read_reg(e, REG_EAX, 3);
        case X86_REG_AX:  return emu_read_reg(e, REG_EAX, 2);
        case X86_REG_AH:  return emu_read_reg(e, REG_EAX, 1);
        case X86_REG_AL:  return emu_read_reg(e, REG_EAX, 0);

        case X86_REG_ECX: return emu_read_reg(e, REG_ECX, 3);
        case X86_REG_CX:  return emu_read_reg(e, REG_ECX, 2);
        case X86_REG_CH:  return emu_read_reg(e, REG_ECX, 1);
        case X86_REG_CL:  return emu_read_reg(e, REG_ECX, 0);

        case X86_REG_EDX: return emu_read_reg(e, REG_EDX, 3);
        case X86_REG_DX:  return emu_read_reg(e, REG_EDX, 2);
        case X86_REG_DH:  return emu_read_reg(e, REG_EDX, 1);
        case X86_REG_DL:  return emu_read_reg(e, REG_EDX, 0);

        case X86_REG_EBX: return emu_read_reg(e, REG_EBX, 3);
        case X86_REG_BX:  return emu_read_reg(e, REG_EBX, 2);
        case X86_REG_BH:  return emu_read_reg(e, REG_EBX, 1);
        case X86_REG_BL:  return emu_read_reg(e, REG_EBX, 0);

        case X86_REG_ESP: return emu_read_reg(e, REG_ESP, 3);
        case X86_REG_SP:  return emu_read_reg(e, REG_ESP, 2);

        case X86_REG_EBP: return emu_read_reg(e, REG_EBP, 3);
        case X86_REG_BP:  return emu_read_reg(e, REG_EBP, 2);

        case X86_REG_ESI: return emu_read_reg(e, REG_ESI, 3);
        case X86_REG_SI:  return emu_read_reg(e, REG_ESI, 2);

        case X86_REG_EDI: return emu_read_reg(e, REG_EDI, 3);
        case X86_REG_DI:  return emu_read_reg(e, REG_EDI, 2);

        default: return 0;
    }
}

static uint32_t emu_write_reg_cs(struct emu *e, x86_reg n, uint32_t word)
{
    switch (n)
    {
        case X86_REG_EAX: return emu_write_reg(e, REG_EAX, 3, word);
        case X86_REG_AX:  return emu_write_reg(e, REG_EAX, 2, word);
        case X86_REG_AH:  return emu_write_reg(e, REG_EAX, 1, word);
        case X86_REG_AL:  return emu_write_reg(e, REG_EAX, 0, word);

        case X86_REG_ECX: return emu_write_reg(e, REG_ECX, 3, word);
        case X86_REG_CX:  return emu_write_reg(e, REG_ECX, 2, word);
        case X86_REG_CH:  return emu_write_reg(e, REG_ECX, 1, word);
        case X86_REG_CL:  return emu_write_reg(e, REG_ECX, 0, word);

        case X86_REG_EDX: return emu_write_reg(e, REG_EDX, 3, word);
        case X86_REG_DX:  return emu_write_reg(e, REG_EDX, 2, word);
        case X86_REG_DH:  return emu_write_reg(e, REG_EDX, 1, word);
        case X86_REG_DL:  return emu_write_reg(e, REG_EDX, 0, word);

        case X86_REG_EBX: return emu_write_reg(e, REG_EBX, 3, word);
        case X86_REG_BX:  return emu_write_reg(e, REG_EBX, 2, word);
        case X86_REG_BH:  return emu_write_reg(e, REG_EBX, 1, word);
        case X86_REG_BL:  return emu_write_reg(e, REG_EBX, 0, word);

        case X86_REG_ESP: return emu_write_reg(e, REG_ESP, 3, word);
        case X86_REG_SP:  return emu_write_reg(e, REG_ESP, 2, word);

        case X86_REG_EBP: return emu_write_reg(e, REG_EBP, 3, word);
        case X86_REG_BP:  return emu_write_reg(e, REG_EBP, 2, word);

        case X86_REG_ESI: return emu_write_reg(e, REG_ESI, 3, word);
        case X86_REG_SI:  return emu_write_reg(e, REG_ESI, 2, word);

        case X86_REG_EDI: return emu_write_reg(e, REG_EDI, 3, word);
        case X86_REG_DI:  return emu_write_reg(e, REG_EDI, 2, word);

        default: return 0;
    }
}

static inline uint32_t emu_get_va(struct emu *e, const x86_op_mem *op)
{
    uint32_t va = op->disp;

    if (op->base != X86_REG_INVALID)
        va += emu_read_reg_cs(e, op->base);

    if (op->index != X86_REG_INVALID)
        va += emu_read_reg_cs(e, op->index) * op->scale;

    return va;
}

static inline int emu_read_mem_cs(struct emu *e, const x86_op_mem *mem,
                                  uint8_t size, uint32_t *o_data)
{
    return emu_read_mem(e, emu_get_va(e, mem), size, o_data);
}

static inline int emu_write_mem_cs(struct emu *e, const x86_op_mem *mem,
                                   uint8_t size, uint32_t data)
{
    return emu_write_mem(e, emu_get_va(e, mem), size, data);
}

static int emu_stack_push(struct emu *e, uint32_t word)
{
    uint32_t p;

    p = e->gpr[REG_ESP] -= 4;

    return emu_write_mem(e, p, 4, word);
}

static int emu_stack_pop(struct emu *e, uint32_t *p_word)
{
    uint32_t p;

    p = e->gpr[REG_ESP];
    e->gpr[REG_ESP] += 4;

    return emu_read_mem(e, p, 4, p_word);
}

static int emu_stack_read(struct emu *e, size_t slot, uint32_t *p_word)
{
    uint32_t p;
    struct seg *s = NULL;

    p = e->gpr[REG_ESP] + slot * sizeof(uint32_t);
    s = emu_find_segment(e, p);

    if (s == NULL)
        return EXIT_FAILURE;

    return seg_read_word(s, p, p_word);
}

static int emu_op_load(struct emu *e, cs_x86_op *src, uint32_t *o_data)
{
    switch (src->type)
    {
    case X86_OP_IMM:
        *o_data = (uint32_t) src->imm;
        return EXIT_SUCCESS;

    case X86_OP_REG:
        *o_data = emu_read_reg_cs(e, src->reg);
        return EXIT_SUCCESS;

    case X86_OP_MEM:
        return emu_read_mem_cs(e, &src->mem, src->size, o_data);

    default:
        return EXIT_FAILURE;
    }
}

static int emu_op_store(struct emu *e, cs_x86_op *dst, uint32_t data)
{
    switch (dst->type)
    {
    case X86_OP_REG:
        emu_write_reg_cs(e, dst->reg, data);
        return EXIT_SUCCESS;

    case X86_OP_MEM:
        return emu_write_mem_cs(e, &dst->mem, dst->size, data);

    default:
        return EXIT_FAILURE;
    }
}

static int emu_disasm(struct emu *e, uint32_t va, cs_x86 **o_insn)
{
    int ret = EXIT_SUCCESS;
    uint8_t *p = NULL;
    const uint8_t *p_insn = NULL;
    uint64_t address = va;
    size_t size = 16;

    if ((ret = emu_get_mem_ptr(e, va, size, &p)) != EXIT_SUCCESS)
        return ret;

    p_insn = p;
    ret = cs_disasm_iter(e->disasm, &p_insn, &size, &address, e->insn);

    if (ret)
        *o_insn = &e->insn->detail->x86;

    return ret ? EXIT_SUCCESS : EXIT_FAILURE;
}

static int emu_find_extern_call(struct emu *e,
                                uint32_t target, struct rel **o_rel)
{
    cs_x86 *insn = NULL;
    uint32_t address = 0;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_disasm(e, target, &insn)) != EXIT_SUCCESS)
        return ret;

    if (e->insn->id != X86_INS_JMP)
        return error("invalid PLT entry: not a jump");

    if (insn->operands[0].type != X86_OP_MEM)
        return error("invalid PLT entry: not a memory location");

    address = (uint32_t) insn->disp;

    if ((*o_rel = emu_find_reloc(e, address)) == NULL)
        return error("invalid external call to %.8x", address);

    return EXIT_SUCCESS;
}

static int emu_exec_insn_mov(struct emu *e, cs_x86 *insn)
{
    uint32_t data;
    int ret = EXIT_SUCCESS;

    if (insn->operands[0].type == X86_OP_REG &&
        (insn->operands[0].reg == X86_REG_CS ||
         insn->operands[0].reg == X86_REG_DS ||
         insn->operands[0].reg == X86_REG_ES ||
         insn->operands[0].reg == X86_REG_FS ||
         insn->operands[0].reg == X86_REG_GS ||
         insn->operands[0].reg == X86_REG_SS))
    {
        e->eip = e->sigill_handler;
        return EXIT_SUCCESS;
    }

    if ((ret = emu_op_load(e, &insn->operands[1], &data)) != EXIT_SUCCESS ||
        (ret = emu_op_store(e, &insn->operands[0], data)) != EXIT_SUCCESS)
    {
        if (e->seg_fail)
        {
            e->seg_fail = 0;
            e->eip = e->sigseg_handler;
            return EXIT_SUCCESS;
        }

        return ret;
    }

    e->eip = e->next_eip;

    return EXIT_SUCCESS;
}

static int emu_handle_sys_exit(struct emu *e)
{
    e->run = 0;
    return EXIT_SUCCESS;
}

static int emu_handle_sys_putchar(struct emu *e)
{
    uint32_t byte = 0;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_stack_read(e, 1, &byte)))
        return ret;

    fputc((char) byte, stdout);
    fflush(stdout);

    return EXIT_SUCCESS;
}

static int emu_handle_sys_puts(struct emu *e)
{
    uint32_t addr = 0;
    uint32_t byte = 0;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_stack_read(e, 1, &addr)))
        return ret;

    for (;;)
    {
        if ((ret = emu_read_mem(e, addr++, 1, &byte)) != EXIT_SUCCESS)
            return ret;

        if (byte == 0)
            break;

        fputc((char) byte, stdout);
    }

    fputc('\n', stdout);;
    fflush(stdout);

    return EXIT_SUCCESS;
}

static int emu_handle_sys_getchar(struct emu *e)
{
    e->gpr[REG_EAX] = getchar();
    return EXIT_SUCCESS;
}

static int emu_handle_sys_malloc(struct emu *e)
{
    uint32_t size = 0;
    struct seg *seg = NULL;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_stack_read(e, 1, &size)))
        return ret;

    if ((seg = emu_add_segment(e, size, e->seg_alloc_base)) == NULL)
        return EXIT_FAILURE;

    e->seg_alloc_base = align(e->seg_alloc_base + size, 16);
    e->gpr[REG_EAX] = seg->base;

    return EXIT_SUCCESS;
}

static int emu_handle_sys_realloc(struct emu *e)
{
    uint32_t addr = 0;
    uint32_t size = 0;
    struct seg *old_seg;
    struct seg *new_seg;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_stack_read(e, 1, &addr)) ||
        (ret = emu_stack_read(e, 2, &size)))
    {
        return ret;
    }

    if ((old_seg = emu_find_segment(e, addr)) == NULL)
        return error("guest failed");

    if ((new_seg = emu_add_segment(e, size, e->seg_alloc_base)) == NULL)
        return EXIT_FAILURE;

    memcpy(new_seg->ptr, old_seg->ptr, MIN(old_seg->size, new_seg->size));
    list_remove(&old_seg->node);
    emu_free_seg(old_seg);

    e->seg_alloc_base = align(e->seg_alloc_base + size, 16);
    e->gpr[REG_EAX] = new_seg->base;

    return EXIT_SUCCESS;
}

static int emu_handle_sys_strlen(struct emu *e)
{
    uint32_t len = 0;
    uint32_t addr = 0;
    uint32_t byte = 0;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_stack_read(e, 1, &addr)))
        return ret;

    for (;;)
    {
        if ((ret = emu_read_mem(e, addr++, 1, &byte)) != EXIT_SUCCESS)
            return ret;

        if (byte == 0)
            break;

        len++;
    }

    e->gpr[REG_EAX] = len;

    return EXIT_SUCCESS;
}

static int emu_handle_sys_call(struct emu *e, const char *name)
{
    static struct handler
    {
        const char *name;
        int (* execute)(struct emu *e);
    }
    handlers[] =
    {
        { "exit",    emu_handle_sys_exit    },
        { "putchar", emu_handle_sys_putchar },
        { "getchar", emu_handle_sys_getchar },
        { "puts",    emu_handle_sys_puts },
        { "malloc",  emu_handle_sys_malloc  },
        { "realloc", emu_handle_sys_realloc },
        { "strlen",  emu_handle_sys_strlen },
    };

    for (size_t i = 0; i < sizeof(handlers) / sizeof(handlers[0]); i++)
    {
        struct handler *h = &handlers[i];

        if (!strcmp(name, h->name))
            return h->execute(e);
    }

    return error("unknown system call: %s", name);
}

static int emu_exec_insn_jmp(struct emu *e, cs_x86 *insn)
{
    uint32_t target = 0;
    struct rel *rel = NULL;
    int ret = EXIT_SUCCESS;

    if (insn->operands[0].type != X86_OP_MEM)
        return EXIT_FAILURE;

    if ((ret = emu_op_load(e, &insn->operands[0], &target)) ||
        (ret = emu_find_extern_call(e, target, &rel)) ||
        (ret = emu_handle_sys_call(e, rel->name)))
    {
         return EXIT_FAILURE;
    }

    return emu_stack_pop(e, &e->eip);
}

static int emu_handle_sigaction(struct emu *e)
{
    uint32_t signum = 0;
    uint32_t sigact = 0;
    uint32_t handler = 0;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_stack_read(e, 1, &signum)) ||
        (ret = emu_stack_read(e, 2, &sigact)) ||
        (ret = emu_read_mem(e, sigact, 4, &handler)))
    {
        return ret;
    }

    msg("sigaction: signal=%-2i handler=%.8x", signum, handler);

    if (signum == 11)
        e->sigseg_handler = handler;
    else if (signum == 4)
        e->sigill_handler = handler;
    else
        return error("unknown signal: %i", signum);

    return EXIT_SUCCESS;
}

static int emu_handle_extern_call(struct emu *e, uint32_t target)
{
    cs_x86 *insn = NULL;
    struct rel *rel = NULL;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_disasm(e, target, &insn)) != EXIT_SUCCESS ||
        (ret = emu_find_extern_call(e, target, &rel)))
    {
        return ret;
    }

    if (strcmp(rel->name, "sigaction"))
        return error("unknown external call: %s", rel->name);

    return emu_handle_sigaction(e);
}

static int emu_exec_insn_call(struct emu *e, cs_x86 *insn)
{
    uint32_t target = 0;
    int ret = EXIT_SUCCESS;

    if (insn->operands[0].type != X86_OP_IMM)
        return EXIT_FAILURE;

    if ((ret = emu_stack_push(e, e->next_eip)) != EXIT_SUCCESS)
        return ret;

    target = (uint32_t) insn->operands[0].imm;

    if ((ret = emu_handle_extern_call(e, target)) != EXIT_SUCCESS)
        return ret;

    return emu_stack_pop(e, &e->eip);
}

static int emu_exec_insn(struct emu *e)
{
    cs_x86 *insn = NULL;
    int ret = EXIT_SUCCESS;

    if ((ret = emu_disasm(e, e->eip, &insn)) == EXIT_SUCCESS)
    {
        cs_x86 *x86_insn = &e->insn->detail->x86;

        e->next_eip = e->eip + e->insn->size;

        switch (e->insn->id)
        {
            case X86_INS_MOV:
                return emu_exec_insn_mov(e, x86_insn);

            case X86_INS_JMP:
                return emu_exec_insn_jmp(e, x86_insn);

            case X86_INS_CALL:
                return emu_exec_insn_call(e, x86_insn);

            default:
                return error("unknown instruction `%s'", e->insn->mnemonic);
        }
    }

    return error("disassembler failed: %s", cs_strerror(cs_errno(e->disasm)));
}

static int emu_exec(struct emu *e)
{
    int ret = EXIT_SUCCESS;

    e->run = 1;

    while (e->run)
    {
        if ((ret = emu_exec_insn(e)) != EXIT_SUCCESS)
            return error("execution failed");
    }

    return EXIT_SUCCESS;
}

static void dump_elf(struct elf *elf)
{
    int i = 0;
    const Elf32_Phdr *ph = NULL;
    const Elf32_Shdr *sh = NULL;

    printf("Segments    type RWX start    end      size\n");
    printf("-----------------------------------------------\n");

    while (elf_enum_segments(elf, &i, &ph))
    {
        printf(" %2i | %10i %c%c%c %.8x %.8x %u\n",
               i - 1,
               ph->p_type,
               "-R"[!!(ph->p_flags & PF_R)],
               "-W"[!!(ph->p_flags & PF_W)],
               "-X"[!!(ph->p_flags & PF_X)],
               ph->p_vaddr, ph->p_vaddr + ph->p_memsz,
               ph->p_memsz);
    }

    printf("\n");

    printf("Sections     type flags    link address  offset   size     name\n");
    printf("-------------------------------------------------------------------\n");

    while (elf_enum_sections(elf, &i, &sh))
    {
        printf(" %2i |  %10i %-8.2x %-4u %.8x %.8x %-8u %s\n",
               i - 1,
               sh->sh_type, sh->sh_flags, sh->sh_link,
               sh->sh_addr, sh->sh_offset, sh->sh_size,
               elf_get_string(elf, elf->sh_strtab, sh->sh_name));
    }

    printf("\n");

#if 0
    printf("Symbols  sh  info size  value    name\n");
    printf("-----------------------------------------\n");

    while (elf_enum_sections(elf, &i, &sh))
    {
        if (sh->sh_type == SHT_DYNSYM)
        {
            int j = 0;
            const Elf32_Sym *sym;

            while (elf_enum_symbols(elf, sh, &j, &sym))
            {
                printf(" %2i |    %-3u %.2x   %-5u %-8x %s\n",
                       j - 1, i - 1,
                       sym->st_info, sym->st_size, sym->st_value,
                       elf_get_string(elf, sh->sh_link, sym->st_name));
            }
        }
    }

    printf("\n");
#endif
    (void) elf_enum_symbols;

    printf("Relocations  offset  type value    addend   name\n");
    printf("----------------------------------------------------\n");

    while (elf_enum_sections(elf, &i, &sh))
    {
        if (sh->sh_type == SHT_REL || sh->sh_type == SHT_RELA)
        {
            int j = 0;
            Elf32_Rela re;

            while (elf_enum_reloc(elf, sh, &j, &re))
            {
                size_t i_sym = ELF32_R_SYM(re.r_info);
                const Elf32_Shdr *symtab = elf_get_section(elf, sh->sh_link);
                const Elf32_Sym *sym = elf_get_symbol(elf, symtab, i_sym);

                printf(" %2i |        %.8x %3u %-8x %-8x %s\n",
                       j - 1,
                       re.r_offset,
                       ELF32_R_TYPE(re.r_info),
                       sym->st_value,
                       re.r_addend,
                       elf_get_string(elf, symtab->sh_link, sym->st_name));
            }
        }
    }

    printf("\n");

    printf("Entry point: %.8x\n", elf->entry);
    printf("\n");
}

static int read_file(const char *path, uint8_t **p_buf, size_t *n_buf)
{
    long size;
    FILE *stream = fopen(path, "rb");
    void *buf;

    if (stream == NULL)
        return EXIT_FAILURE;

    fseek(stream, 0, SEEK_END);
    size = ftell(stream);
    fseek(stream, 0, SEEK_SET);

    if ((buf = malloc((size_t) size)) == NULL)
    {
        fclose(stream);
        return EXIT_FAILURE;
    }

    fread(buf, 1, size, stream);

    *p_buf = buf;
    *n_buf = (size_t) size;

    fclose(stream);

    return EXIT_SUCCESS;
}

int main(int argc, const char **argv)
{
    uint8_t *p_elf = NULL;
    size_t n_elf = 0;
    struct elf elf;
    struct emu emu;
    int ret = EXIT_SUCCESS;

    if (argc != 2)
    {
        printf("USAGE: linemu ELFFILE\n");
        printf("\n");
        printf(" (c) 2020 by Microedition. All right reserved and reversed\n");
        printf("\n");
        return EXIT_SUCCESS;
    }

    if ((ret = read_file(argv[1], &p_elf, &n_elf)) != EXIT_SUCCESS)
        return error("failed to open %s", argv[1]);

    ret = elf_parse(&elf, p_elf, n_elf);

    if (ret != EXIT_SUCCESS)
    {
        error("elf_parse() failed");
        goto cleanup;
    }

    dump_elf(&elf);
    msg("loading...");

    if ((ret = emu_init(&emu, &elf)) != EXIT_SUCCESS)
    {
        error("emu_init() failed");
        goto cleanup;
    }

    if ((ret = emu_stack_push(&emu, 0x00000000)) ||
        (ret = emu_stack_push(&emu, 0x00000000)))
    {
        goto free;
    }

    msg("running...");
    ret = emu_exec(&emu);

    msg("");
    msg("finished with %s", ret == EXIT_SUCCESS ? "success" : "error");
    msg("");

free:
    emu_free(&emu);

cleanup:
    free(p_elf);

    return ret;
}
